function [ varargout ] = SelfSupAM_plot2D( varargin )
%   SelfSupAM_plot2D Plots ARTMAP category boxes and predictions.
%
%   rectangles = SelfSupAM_plot2D( model ) Plots ARTMAP category boxes for 2D problems
%       model ~ java object of long-term memory traces and model parameters
%           returned by SelfSupAM_trainStage#()
%       rectangles ~ rectangle object handles
%
%   [ rectangles, predImage ] = SelfSupAM_plot2D( model, gridPreds ) 
%       Plots ARTMAP category boxes and predictions. 
%       gridPreds ~ n^2x1 integer ~ Vector of predictions that can be
%           reshaped into an image-plottable nxn grid.
%       predImage ~ handle to prediction image
%
%   predImage = SelfSupAM_plot2D( gridPreds ) Plots predictions only. 
%
%   Greg Amis
%   Technology Lab
%   Dept. of Cognitive & Neural Systems
%   Boston University
%   7 November 2008
%
%   Licensing Policy
%   This implementation is made available to the general public under the terms 
%   of Copyleft, as defined by the <a href="http://www.fsf.org/">Free Software Foundation</a> 
%   (see discussion <a href="http://www.gnu.org/copyleft/copyleft.html">here</a> for example). 
%   As such, it is available free of charge, and may be copied, redistributed, or modified, 
%   as long as any resulting work is distributed under the same terms of Copyleft.

if nargin==2
    [model, gridPreds] = varargin{:};
elseif nargin==1
    if isjava(varargin{1})
        model = varargin{1};
        gridPreds = [];
    else
        model = [];
        gridPreds =  varargin{1};
    end;
else
    error('Wrong number of input arguments: %d',nargin);
end;

if isempty(model)
    L = max(gridPreds);
else
    L = model.getL();
end;
    

colors = [ 1 0 0 ; 0 0 1; 0 1 0; 0 1 1; 1 0 1; 1 1 0; 0.25 0.25 0.25; 0 0 0; ];
boxLineWidth = 1.5;
ptBoxSize = 0.01;
lineBoxSize = 0.002;

if size(colors,1)<L
    error('Insufficient colors for number of classes');
end;

if ~isempty(gridPreds)
    
    %Reshape preds into nxn grid
    sideLength = sqrt(length(gridPreds));
    if sideLength~=round(sideLength)
        error('Predictions not square');
    end;
    predSquare = reshape(gridPreds,sideLength,sideLength);

    %Convert to nxnx3 color matrix
    Cr = zeros(size(predSquare));
    Cg = Cr; Cb = Cr;

    for k = 1:L
        color = colors(k,:);
        isK = predSquare==k;
        Cr( isK ) = color(1);
        Cg( isK ) = color(2);
        Cb( isK ) = color(3);
    end;

    C(:,:,1) = Cr;
    C(:,:,2) = Cg;
    C(:,:,3) = Cb;
    
    predImage = image([0 1],[0 1],C);   
    colors(:) = 0; %Make colors all black when plotting both preds and boxes
else
    predImage = [];
end;

if ~isempty(model)

    % Plot boxes
    w = model.w()';
    [twoM C] = size(w);
    u = w(1:2,:);
    v = 1 - w(3:4,:);
    sizeRj = abs(v-u);

    %Make point boxes and line boxes visible
    sizeRj(:,all(sizeRj<ptBoxSize)) = ptBoxSize;
    sizeRj(sizeRj<lineBoxSize) = lineBoxSize;


    hold on;
    rectangles = zeros(1,C);
    kappa = model.kappa() + 1;
    undercommitted = v < u;
    
    for j = 1:C
        
        rectangles(j) = rectangle( 'Position',[ min(u(:,j),v(:,j)); sizeRj(:,j) ],...
            'EdgeColor',colors(kappa(j),:)*0.7,...
            'LineWidth',boxLineWidth );
        
        undercommittedFeature = find(undercommitted(:,j));
        if ~isempty(undercommittedFeature)
            lineFeature = find(sizeRj(:,j)<=lineBoxSize+1e-7);
            if ~isempty(lineFeature) 
                %Linebox. Can't just make linestyle dashed, because the two
                %rectangle edges overlap and make the line look solid.
                delete(rectangles(j));
                if lineFeature==1
                    X = repmat( mean( [u(1,j), v(1,j)] ), 2, 1 );
                    Y = [ v(2,j); u(2,j) ];
                else % lineFeature==2
                    X = [ v(1,j); u(1,j) ]; 
                    Y = repmat( mean( [u(2,j), v(2,j)] ), 2, 1 );
                end;
                line(X,Y,'Color',colors(kappa(j),:)*0.7,'LineWidth',boxLineWidth,'LineStyle','--');
            else
                set(rectangles(j),'LineStyle','--');
            end;
        end;
        if length(undercommittedFeature)==1 %Other feature is committed, add in solid lines
            if undercommittedFeature==1
                X = [u([1,1],j),v([1,1],j)];
                Y = repmat([u(2,j);v(2,j)],1,2);
            else %undercommittedFeature==2
                X = repmat([u(1,j);v(1,j)],1,2);
                Y = [u([2,2],j),v([2,2],j)];
            end;
            line(X,Y,'Color',colors(kappa(j),:)*0.7,'LineWidth',boxLineWidth);
        end;
    end;
end;

axis square;
axis xy;
%axis( [ 0 1 0 1 ] );

if nargout==2
    varargout = { rectangles, predImage };
elseif nargout==1
    if ~isempty(model)
        varargout = {rectangles};
    else
        varargout = {predImage};
    end;
else
    %Do nothing.
end;