function model = SelfSupAM_trainStage1( varargin )
%   SelfSupAM_trainStage1 Performs Stage 1 training using Self-Supervised ARTMAP.
%
%   model = SelfSupAM_trainStage1( inputs, labels ) 
%     trains Self-Supervised ARTMAP with default parameters,
%           using the passed training inputs and training class labels.
%       inputs ~ MxN floating point (double|single), 
%           where N is the number of inputs, and all values are already
%           scaled to the unit interval: all(inputs(:)>=0 & inputs(:)<=1)
%       labels ~ Nx1 integer ~ ground-truth labels
%
%   model = SelfSupAM_trainStage1( inputs, labels, model ) 
%     trains Self-Supervised ARTMAP using the passed model, 
%     training inputs and training class labels. Typically, one would use 
%     this invocation to update an already trained model or train a new model 
%     when it's more convenient to pass the parameters as a model object rather 
%     than as a sequence of name:value pairs.
%       model ~ Java object of long-term memory traces and model parameters
%           Can be just an untrained model instance, in which case new memory
%           traces will be created. Valid parameters are listed below.
%
%   model = SelfSupAM_trainStage1( inputs, labels, 'PropertyName',propertyvalue ) 
%     trains Self-Supervised ARTMAP using the passed inputs, labels, and
%     then parameters as name:value pairs. Valid parameters are below. 
%     All are optional and, if not specified, are either inferred or 
%     take the default value defined in Amis & Carpenter (2009).
%     Parameters
%       M ~ Number of input features
%       L ~ Number of output classes
%       V ~ Number of voters (defaults to V=1)
%       classCap ~ Maximum number of points from each class
%       alpha ~ signal rule parameter
%       beta1 ~ Stage 1 learning fraction
%       beta2 ~ Stage 2 learning fraction
%       epsilon ~ match tracking parameter
%       rhobar ~ baseline vigilance
%       p ~ CAM rule power
%       E1 ~ # stage1 training epochs
%       E2 ~ # stage2 training epochs
%   
%           
%   Based on:   Amis, G., & Carpenter, G.A. (2009). 
%                   Self-Supervised ARTMAP. (in preparation)
%   
%
%   Greg Amis
%   Technology Lab
%   Dept. of Cognitive & Neural Systems
%   Boston University
%   26 March 2009
%
%   Licensing Policy
%   This implementation is made available to the general public under the terms 
%   of Copyleft, as defined by the <a href="http://www.fsf.org/">Free Software Foundation</a> 
%   (see discussion <a href="http://www.gnu.org/copyleft/copyleft.html">here</a> for example). 
%   As such, it is available free of charge, and may be copied, redistributed, or modified, 
%   as long as any resulting work is distributed under the same terms of Copyleft.

if nargin<2, error('Inputs and labels required.'); end;

inputs = varargin{1};
labels = varargin{2};

SelfSupAM_loadJava();

if nargin==3
    % Presume SelfSupAM_trainStage1( inputs, labels, model )
    model = varargin{3};
else
    %Create model with default parameters
    params = {'V',1};
    if nargin>3
        params = [params,varargin(3:end)];
    end;
    model = SelfSupAM_createInstance(params{:});
end;
    
M = model.getNumberOfFeatures();
if M==-1
    M = size(inputs,1);
    model.setNumberOfFeatures(M);
end;

L = model.getNumberOfClasses();
if L==-1
    L = max(labels); 
    model.setNumberOfClasses(L);
end;

labeledInputs = edu.bu.cns.data.LabeledInputsImpl( inputs', int32(labels-1), L);
model.setUp();

model.trainStage1(labeledInputs);

