function [ cmat ] = conf_mat( actuals, predictions, classNames )
%   conf_mat Generates (and optionally plots) a confusion matrix.
%
%   conf_mat( actuals, predictions, classNames ) Generates and plots the confusion matrix 
%       actuals ~ 1xN vector of actual class labels
%       predictions ~ 1xN vector of predicted labels
%       classNames ~ 1xL cell of names for each class 
%
%   [ cmat ] = conf_mat( actuals, predictions ) Generates and returns 
%       the confusion matrix without plotting it. 
%       cmat ~ NxN matrix, with actuals on the rows and predictions on the columns
%
%   Greg Amis
%   Technology Lab
%   Dept. of Cognitive & Neural Systems
%   Boston University
%   30 March 2009
%
%   Licensing Policy
%   This implementation is made available to the general public under the terms 
%   of Copyleft, as defined by the <a href="http://www.fsf.org/">Free Software Foundation</a> 
%   (see discussion <a href="http://www.gnu.org/copyleft/copyleft.html">here</a> for example). 
%   As such, it is available free of charge, and may be copied, redistributed, or modified, 
%   as long as any resulting work is distributed under the same terms of Copyleft.

if nargin~=3, error('All arguments required.'); end;

L = length(classNames);
cmat = zeros(L);
boxEdgeColor = 0.5*ones(1,3);

for k = 1:L
    actualk = actuals==k;
    predk = predictions(actualk);
    for kp = 1:L
        cmat(k,kp) = sum( predk==kp );
    end;
end;

if nargout==1
    return;
end;

%Plot

xlabel('Predicted Class','FontSize',10);
ylabel('Actual Class','FontSize',10);

classNames{end+1} = '%';
set(gca, 'FontSize',10,'FontAngle','italic',...
    'XAxisLocation','top', 'YDir','reverse',...
    'XLim',[ 0.5 L+1.5 ], 'XTick',1:L+1, 'XTickLabel',classNames,...
    'YLim',[ 0.5 L+1.5 ], 'YTick',1:L, 'YTickLabel',classNames(1:end-1));

hold on;

for k = 1:L
    for kp = 1:L
        rectangle( 'Position',[kp-.5,k-.5,1,1],...
            'EdgeColor',boxEdgeColor, 'FaceColor',[1 1 1] );
        c = cmat(k,kp);
        if c==0, str = ''; else str = sprintf('%d',c); end;
        t = text( kp, k, str, 'FontSize',10,...
            'HorizontalAlignment','center', 'VerticalAlignment','middle');
        if k==kp, set(t,'FontWeight','bold'); end;
    end;
end;

classAccs = NaN(1,L);
for k = 1:L

    rectangle( 'Position',[L+.5,k-.5,1,1],...
        'EdgeColor',boxEdgeColor, 'FaceColor',[1 1 1] );
    classAccs(k) = cmat(k,k) / sum( cmat(k,:) );
    text( L+1.4,k, sprintf('%1.1f%%', 100*classAccs(k)), 'FontSize',10,...
            'HorizontalAlignment','right', 'VerticalAlignment','middle');
    
end;

rectangle( 'Position',[.5,L+.5,L,1], 'EdgeColor',boxEdgeColor, 'FaceColor',[1 1 1] );
text( L+.4,L+1, 'Class-based accuracy', 'FontSize',10,...
        'HorizontalAlignment','right', 'VerticalAlignment','middle');

rectangle( 'Position',[L+.5,L+.5,1,1], 'EdgeColor',boxEdgeColor, 'FaceColor',[1 1 1] );
totalAcc = mean(classAccs);
text( L+1.4,L+1, sprintf('%1.1f%%', 100*totalAcc),...
    'FontSize',10, 'FontWeight','bold',...
    'HorizontalAlignment','right', 'VerticalAlignment','middle');

