/**
 *	SelfSupCrossValidator.java
 *
 *  Greg Amis
 *  Technology Lab
 *  Dept. of Cognitive & Neural Systems
 *  Boston University
 *  March 26, 2009
 *
 *  Licensing Policy
 *  This implementation is made available to the general public under the terms 
 *  of Copyleft, as defined by the <a href="http://www.fsf.org/">Free Software Foundation</a> 
 *  (see discussion <a href="http://www.gnu.org/copyleft/copyleft.html">here</a> for example). 
 *  As such, it is available free of charge, and may be copied, redistributed, or modified, 
 *  as long as any resulting work is distributed under the same terms of Copyleft.
 */
package edu.bu.cns.classifiers;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.concurrent.Callable;

import edu.bu.cns.classifiers.SelfSupervisedClassifier;
import edu.bu.cns.data.LabeledInput;
import edu.bu.cns.data.LabeledInputs;
import edu.bu.cns.data.LabeledInputsImpl;
import edu.bu.cns.util.ForkProcessor;

/**
 *	Evaluates a self-supervised learning system 
 *	via cross-validation on a benchmark dataset.
 *	Specifically, it first splits a dataset, 
 *		consisting of labeled and unlabeled points, into F disjoint strips.
 *	Then, for each strip f:
 *	  1. Performs Stage 1 training using a feature subset of labeled points from the F-1 other strips.
 *	  2. Performs Stage 2 training using all features for all points from strip f.
 *	  3. Tests using all features for the labeled points in strip f.
 *
 *	Uses {@link ForkProcessor} to train and test for each strip in parallel, 
 *	assuming a multi-core machine.
 */
public class SelfSupCrossValidator 
implements Callable<SelfSupCrossValidator.Result> 
{
	
	protected final SelfSupervisedClassifier model;
	protected final LabeledInputs[] srcStrips;
	protected final int[] stage1FeatureIndexes;
	
	protected LabeledInputsImpl[] stage1Strips;
	
	public SelfSupCrossValidator( 
			final SelfSupervisedClassifier model,
			final LabeledInputs[] srcStrips, 
			int[] stage1FeatureIndexes ) {
		this.model = model;
		this.srcStrips = srcStrips;
		this.stage1FeatureIndexes = stage1FeatureIndexes;
	}

	/** 
	 * Performs cross validation and returns the results.
	 * See class description for details.
	 * @return
	 * @throws Exception
	 */
	public Result call() throws Exception {
		
		createStage1Strips();
		
		final int F = srcStrips.length;
		final ForkProcessor<StripResult> fp = new ForkProcessor<StripResult>(F);
		
		for (int f = 0; f<F; f++)
			fp.submit( new StripExerciser(f) );
		
		if (fp.isCanceled()) 
			return null;
		
		final List<StripResult> stripResults = fp.finish();
		final Result result = new Result(
				new ArrayList<int[]>(F), new ArrayList<int[]>(F),
				new ArrayList<SelfSupervisedClassifier>(F) );
		
		for (final StripResult sr : stripResults) {
			result.predsAfterStage1.add(sr.predsAfterStage1);
			result.predsAfterStage2.add(sr.predsAfterStage2);
			result.modelsAfterStage2.add(sr.modelAfterStage2);
		}
		
		return result;
	}
	
	/** Creates stage 1 strips, which contain only labeled inputs 
	 * and only features in stage1FeatureIndexes
	 */
	private void createStage1Strips() {
		
		final int M = srcStrips[0].getNumberOfFeatures();
		final int L = srcStrips[0].getNumberOfClasses();
		stage1Strips = new LabeledInputsImpl[srcStrips.length];
		
		/** Vector with all features missing. */
		final float[] nanFeatures = new float[M];
		Arrays.fill(nanFeatures,Float.NaN);
		
		for (int f = 0; f<srcStrips.length; f++) {
			
			final LabeledInputs srcStrip = srcStrips[f];
			final List<LabeledInput> destInputs = new ArrayList<LabeledInput>(256);
			
			for (final LabeledInput input : srcStrip) {
				if (input.isLabeled()) {
					final float[] srcFeatures = input.features();
					final float[] dstFeatures = copy(nanFeatures);
					for (final int i : stage1FeatureIndexes)
						dstFeatures[i] = srcFeatures[i];		
					destInputs.add( new LabeledInput.Impl(dstFeatures,input.label()) );
				}
			}
			
			stage1Strips[f] = new LabeledInputsImpl(destInputs,L);
			
		}
		
	}
	
	/** Returns a copy of the passed array */
	private static final float[] copy( final float[] src ) {
		final float[] c = new float[src.length];
		System.arraycopy(src, 0, c, 0, src.length);
		return c;
	}
	
	
	/**
	 * Exercises model for particular strip.
	 */
	private class StripExerciser implements Callable<StripResult> {
		
		/** Strip index for Stage 2 and testing */
		private final int targetStripIndex; 
		
		private StripExerciser(final int targetStripIndex) { 
			this.targetStripIndex = targetStripIndex; 
		}
		
		public StripResult call() {
			
			final LabeledInputs targetStrip = srcStrips[targetStripIndex];
			final int L = targetStrip.getNumberOfClasses();
			final int F = srcStrips.length;

			
			//Make Stage 1 training from all inputs and only stage1FeatureIndexes
			// from all other strips.
			
			//First, count how large Stage 1 training set is
			int Ns1 = 0;
			for (int f = 0; f<F; f++) {
				if (f!=targetStripIndex) 
					Ns1 += stage1Strips[f].size();
			}
			
			//Then, join all the other strips together.
			final List<LabeledInput> s1Inputs = new ArrayList<LabeledInput>(Ns1);
			for (int f = 0; f<F; f++) {
				if (f!=targetStripIndex) 
					s1Inputs.addAll(stage1Strips[f].list());
			}
			final LabeledInputs stage1Training = new LabeledInputsImpl(s1Inputs,L);
			
			//Stage 2 training is just the targetStrip
			
			//Make test set from labeled points in targetStrip.
			final int Ntest = stage1Strips[targetStripIndex].size();
			final List<LabeledInput> testInputs = new ArrayList<LabeledInput>(Ntest);
			for (final LabeledInput input : targetStrip) {
				if (input.isLabeled())
					testInputs.add(input);
			}							
			final LabeledInputs testing = new LabeledInputsImpl(testInputs,L);
			
			//Clone model, so multiple threads don't try 
			// to train on the same model at the same time
			final SelfSupervisedClassifier stripModel = model.clone();
			stripModel.setUp();
			
			stripModel.trainStage1(stage1Training);
			final int[] s1Preds = stripModel.test(testing);
			stripModel.trainStage2(targetStrip);
			final int[] s2Preds = stripModel.test(testing);
			
			return new StripResult(s1Preds,s2Preds,stripModel);
		}
		
	}
	
	private class StripResult {
		private final int[] predsAfterStage1, predsAfterStage2;
		private final SelfSupervisedClassifier modelAfterStage2;
		public StripResult(
				int[] predsAfterStage1, int[] predsAfterStage2,
				SelfSupervisedClassifier modelAfterStage2) {
			this.predsAfterStage1 = predsAfterStage1;
			this.predsAfterStage2 = predsAfterStage2;
			this.modelAfterStage2 = modelAfterStage2;
		}
		
	}
	
	
	public static class Result {
		
		public final List<int[]> predsAfterStage1, predsAfterStage2;
		public final List<SelfSupervisedClassifier> modelsAfterStage2;
		
		public Result(
				List<int[]> predsAfterStage1,
				List<int[]> predsAfterStage2,
				List<SelfSupervisedClassifier> modelsAfterStage2) {
			this.predsAfterStage1 = predsAfterStage1;
			this.predsAfterStage2 = predsAfterStage2;
			this.modelsAfterStage2 = modelsAfterStage2;
		}
		
	}
	

}
