/**
 *	SupervisedClassifier.java
 *
 *  Greg Amis
 *  Technology Lab
 *  Dept. of Cognitive & Neural Systems
 *  Boston University
 *  Nov 13, 2008
 *
 *  Licensing Policy
 *  This implementation is made available to the general public under the terms 
 *  of Copyleft, as defined by the <a href="http://www.fsf.org/">Free Software Foundation</a> 
 *  (see discussion <a href="http://www.gnu.org/copyleft/copyleft.html">here</a> for example). 
 *  As such, it is available free of charge, and may be copied, redistributed, or modified, 
 *  as long as any resulting work is distributed under the same terms of Copyleft.
 */
package edu.bu.cns.classifiers;

import edu.bu.cns.data.Input;
import edu.bu.cns.data.Inputs;
import edu.bu.cns.data.LabeledInputs;

/**
 *	Interface for models that can learn classification patterns from supervised training.
 */
public interface SupervisedClassifier {
	
	/**
	 * To be called after all free parameters are set,
	 * before any trainning.
	 */
	public void setUp();

	/**
	 * Trains the classifier using the passed labeled inputs
	 * @param labeledInputs
	 */
	public void train(LabeledInputs labeledInputs);
	
	/**
	 * Tests the classifier on a single input.
	 * @param input
	 * @return class prediction
	 */
	public int test(Input input);
	
	/**
	 * Outputs generated by the classifier are added to Outputs
	 * @param inputs
	 * @return class predictions
	 */
	public int[] test(Inputs<?> inputs);
	
	/**
	 * Tests the classifier on a single input.
	 * @param input
	 * @return p(k|input)
	 */
	public float[] testDistribution(Input input);
	
	/**
	 * Outputs generated by the classifier are added to Outputs
	 * @param inputs
	 * @return class probabilities
	 */
	public float[][] testDistribution(Inputs<?> inputs);
	
	/**
	 * Trains and tests, returning class predictions.
	 *
	 * @param trainingInputs
	 * @param testInputs
	 * @return class predictions
	 */
	public int[] exercise( LabeledInputs trainingInputs, Inputs<?> testInputs);
	
	/**
	 * Trains and tests, returning class probabilities.
	 * 
	 * @param trainingInputs
	 * @param selfSupInputs
	 * @param testInputs
	 * @return class probabilities
	 */
	public float[][] exerciseDistribution( LabeledInputs trainingInputs, Inputs<?> testInputs);
	
	/**
	 * Creates a deep copy of the classifier.
	 * @return
	 */
	public SupervisedClassifier clone();
	
	public int getNumberOfFeatures();
	public void setNumberOfFeatures(int M);
	
	public int getNumberOfClasses();
	public void setNumberOfClasses(int L);
	
}
