/**
 *	SelfSupervisedARTMAP.java
 *
 *  Greg Amis
 *  Technology Lab
 *  Dept. of Cognitive & Neural Systems
 *  Boston University
 *  March 25, 2009
 *  
 *  Licensing Policy
 *  This implementation is made available to the general public under the terms 
 *  of Copyleft, as defined by the <a href="http://www.fsf.org/">Free Software Foundation</a> 
 *  (see discussion <a href="http://www.gnu.org/copyleft/copyleft.html">here</a> for example). 
 *  As such, it is available free of charge, and may be copied, redistributed, or modified, 
 *  as long as any resulting work is distributed under the same terms of Copyleft.
 */
package edu.bu.cns.classifiers.artmap;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.NoSuchElementException;
import java.util.Random;

import edu.bu.cns.classifiers.AbsSelfSupClassifier;
import edu.bu.cns.classifiers.ClassLabelOutOfBoundsException;
import edu.bu.cns.classifiers.IllegalParameterException;
import edu.bu.cns.classifiers.LabelRequiredException;
import edu.bu.cns.classifiers.VoterEnsemble;
import edu.bu.cns.data.Input;
import edu.bu.cns.data.InputFeaturesLengthMismatchException;
import edu.bu.cns.data.Inputs;
import edu.bu.cns.data.LabeledInput;
import edu.bu.cns.data.LabeledInputs;
import edu.bu.cns.util.FloatVector;
import edu.bu.cns.util.IntArrayList;
import edu.bu.cns.util.Pow;

/**
 * Implementation of Self-Supervised ARTMAP, as described in:
 * 	Amis, G., & Carpenter, G.A. (2009). Self-Supervised ARTMAP. (in preparation)
 * 
 * This class does not do cross-validation or voting. Please see {@link CrossValidation}
 * and {@link VoterEnsemble} for those features and SSART_benchmarks.m for example usage.
 * 
 */
public class SelfSupervisedARTMAP extends AbsSelfSupClassifier {
	
	private static final long serialVersionUID = -7768128682581338639L;
	
	// PARAMETERS
	
	/** Maximum number of input features */
	protected int M = -1; 
	
	/** Maximum number of output classes */
	protected int L = -1;
	
	/** Number of Stage 1 training epochs */
	protected int E1 = 1;
	
	/** Number of Stage 2 training epochs */
	protected int E2 = -1;
	
	/** Signal rule, alpha in (0,1) */
	protected float alpha = 0.01f;
	
	/** Baseline vigilance, rhobar in [0,1] */
	protected float rhobar = 0f;
	
	/** Learning rate during stage 1 training, beta1 in (0,1] */
	protected float beta1 = 1f;
	
	/** Learning rate during stage 2 learning, beta2 in [0,1] */
	protected float beta2 = 0.01f;
	
	/** Match-tracking parameter, epsilon in (-1,1) */
	protected float epsilon = -0.01f;
	
	/** CAM rule power, p in (1,Inf] */
	protected float p = 2f;
	
	protected float tieThreshold = 1e-6f;
	protected long tieBreakerSeed = 1l;
	
	// LONG-TERM MEMORY. Recruited coding nodes
	protected List<CodingNode> codingNodes;
	
	
	// TRANSIENTS
	
	/** # features specified in a (2M-|A|) */
	protected transient float Ma;
	
	/** Input field (F0) activation */
	protected transient FloatVector A;	
	
	/** Winning coding node J (WTA) */
	protected transient CodingNode winner;
	
	/** Viglance */
	protected transient float rho;
	
	/** Current learning fraction */
	protected transient float beta;
	
	/** Signal from coding field (F2) to mapping field (Fab)*/
	protected transient float[] sigma;
	
	protected transient List<CodingNode> codingTies;
	protected transient IntArrayList mappingTies;
	
	protected transient Random tieBreaker;
	
	/** Set of active coding nodes */
	protected transient Iterable<CodingNode> Lambda;
	/** Set of point-boxes at a */
	protected transient List<CodingNode> LambdaPrime;
	
	/** Strategy for calculating exponents. */
	protected transient Pow pow;
	
	/**
	 * Initializes all internal variables. 
	 * Should be called after setting all free parameters
	 * and before doing any training.
	 */
	public void setUp() {
		validateParameters();
		A = new FloatVector(2*M);
		if (codingNodes==null)
			codingNodes = new ArrayList<CodingNode>(64*L);
		Lambda = new ActiveCodingNodes();
		LambdaPrime = new ArrayList<CodingNode>(L);
		sigma = new float[L];
		codingTies = new ArrayList<CodingNode>(L);
		mappingTies = new IntArrayList(64*L);
		tieBreaker = new Random(tieBreakerSeed);
		if (E2==-1)
			E2 = Math.round(1f/beta2);
		pow = Pow.createPow(-p);
	}
	
	public void resetLearning() {
		codingNodes.clear();
		if (tieBreaker!=null)
			tieBreaker.setSeed(tieBreakerSeed);
	}
	
	public void trainStage1( final LabeledInputs inputs ) {
		for (int e=0; e<E1; e++) {
			for ( final LabeledInput input : inputs)
				trainStage1( input );
		}
	}
	
	/**
	 * Handles stage 1 training on a single input-class pair (a,K)
	 * @param input labeled input pattern
	 */
	public void trainStage1( final LabeledInput input ) {
		validateLabeledInput(input);
		trainStage1(input.features(), input.label());		
	}
	
	protected void trainStage1( final float[] a, final int K ) {
		
		rho = rhobar;
		beta = beta1;
		
		activateInputField(a);
		inputCodingField();
		
		// Search for the coding node with best match 
		// that passes vigilance and is correct class
		for (;;) {

			chooseWinner();
			
			if (winner==null) {
				recruitNode(K);
				break;
			}
			
			activateCodingFieldWTA();
			
			if ( !passesVigilance() ) {
				deactivate(winner);
				continue;
			}
			
			if ( winner.kappa==K ) {
				resonate(winner);
				break;
			}
			else {
				matchTrack();
				deactivate(winner);
				continue;
			}
		}

	}
	
	public void trainStage2( final Inputs<?> inputs ) {
		for (int e=0; e<E2; e++) {
			for ( final Input input : inputs )
				trainStage2( input );
		}
	}
	
	/**
	 * Handles stage 2 learning on a single input
	 * @param input pattern
	 */
	public void trainStage2( final Input input ) {
		validateInput(input);
		trainStage2(input.features());
	}
	
	protected void trainStage2( final float[] a ) {
		
		beta = beta2;
		
		activateInputField(a);
		inputCodingField();
		activateCodingFieldDist();
		resonate();

	}
	
	/**
	 * Handles testing on a single input (a)
	 * @param a input pattern
	 * @return Approximation of P(k|a) = sigma_k
	 */
	public float[] testDistribution( final Input input ) {
		validateInput(input);
		test(input.features());
		final float[] s = new float[L];
		System.arraycopy(sigma, 0, s, 0, L);
		return s;
	}
	
	protected void test( final float[] a ) {
		
		activateInputField(a);
		inputCodingField();
		activateCodingFieldDist();
		activateMappingField();
		
	}
	
	/**
	 * Handles testing on a single input (a)
	 * @param a input pattern
	 * @return Class label prediction (K')
	 */
	public int test( final Input input ) {
		validateInput(input);
		test(input.features());
		return makePrediction();
	}
	
	/**
	 * Complement-coding:
	 * 	For specified features i: A_i = a_i, A_{M+i} = 1 - a_i 
	 *  For unspecified features i: A_i = A_{M+i} = 1
	 */
	protected void activateInputField(final float[] a) {
		Ma = M;
		//Complement code
		for ( int i=0; i<M; i++ ) {
			if (Float.isNaN(a[i])) { //Missing input feature
				A.set(  i,  1f );
				A.set( M+i, 1f );
				Ma--;
			}
			else { //Existing input feature
				A.set(  i,   a[i]   );
				A.set( M+i, 1f-a[i] );
			}
		}
	}
	
	/**
	 * Choice-by-difference input function,
	 * with support for undercommitted nodes
	 * 
	 * Tj = [ 2M - |A| + |A^w| - |w| ] / [ 1 - (1-alpha)*Phi_j ] - alpha*[ M -|w| ]
	 */
	protected void inputCodingField() {
		
		final float Mf = M;
		final float oneMinusAlpha = 1f-alpha;
		final float Tu = alpha*Mf; //Signal to uncommitted node
		
		for (final CodingNode node : codingNodes) {
			
			final FloatVector w = node.w;
			
			float S = 0f;
			for (int i=0; i<A.size(); i++) {
				S += Math.min( A.get(i), w.get(i) );
			}
			
			node.S = S;
			node.T = ( Ma + S - w.sum() )  / ( 1f - oneMinusAlpha*node.phi )
			  - alpha * ( Mf - w.sum() );
			
			assert node.T > -1e-5f : String.format("T=%g < 0", node.T);;
			assert node.T < Mf + 1e-5f : String.format("T=%g > M=%g", node.T, Mf);
			
			node.isActive = node.T > Tu;
		}
	}
	
	/**
	 * Chooses the active node with the largest input,
	 * returning null if no nodes are active. Breaks ties randomly.
	 * 
	 * @return
	 */
	protected void chooseWinner() {
		
		winner = null;
		float maxT = 0f;
		
		codingTies.clear();
		
		for( final CodingNode node: Lambda ) {
			
			if ( node.T > maxT ) {
				codingTies.clear();
				codingTies.add(node);
				maxT = node.T;
			}
			else if ( maxT-node.T < tieThreshold ) {
				codingTies.add(node);
			}
			
		}
		
		if (codingTies.isEmpty())
			return;
		
		final int tJ = (codingTies.size()==1) ? 0 : tieBreaker.nextInt(codingTies.size()); 
		
		winner = codingTies.get(tJ);
	}
	
	protected void recruitNode(final int K) {
		
		final CodingNode node = new CodingNode();
		node.w = A.clone();
		node.kappa = K;
		node.phi = (M - Ma)/M;
		
		codingNodes.add(node);
	}
	
	/**
	 * @return |A^w|/|A| >= rho
	 */
	protected boolean passesVigilance() {
		return winner.S / A.sum() >= rho;
	}
	
	/**
	 * Raise vigilance to exclude current winning node
	 * and any node with <= match
	 * rho <- |A^w|/|A| + epsilon
	 */
	protected void matchTrack() {
		rho = winner.S / A.sum() + epsilon;
		if (rho<0f) rho = 0f;
		if (rho>1f) rho = 1f;
	}
	
	protected void deactivate(final CodingNode node) {
		node.y = 0f;
		node.isActive = false;
	}
	
	/**
	 * Resonate/Learn on A for j
	 * w_ij -= beta*[ y_j - (1-w_ij) - A_i ]^+
	 */
	protected void resonate(final CodingNode node) {
		final FloatVector w = node.w;
		boolean updatePhi = false;
		for ( int i=0; i<A.size(); i++ ) {
			float dw = node.y - ( 1f - w.get(i) )- A.get(i);
			if ( dw > 0f ) {
				w.addTo(i, -beta*dw);
				updatePhi = true;
			}
		}
		
		if (updatePhi) { 
			float phiM = 0f;
			for ( int i=0; i<M; i++ ) {
				final float phi_ij = w.get(i) + w.get(M+i) - 1f;
				if ( phi_ij > 0f )
					phiM += phi_ij;
			}
			node.phi = phiM/M;
		}
	}
	
	/**
	 * Distributed resonance
	 */
	protected void resonate() {
		for (final CodingNode node : Lambda)
			resonate(node);
	}
	
	protected void activateCodingFieldWTA() {
		//Reset values 
		for (final CodingNode node : codingNodes) 
			node.y = 0f;
		winner.y = 1f;
	}
	
	
	/**
	 * Increased-Gradient CAM Rule
	 * yyj = ( M - T_j )^-p
	 * yj = yyj / sum_k { yyk }
	 */
	protected void activateCodingFieldDist() {

		//Reset values 
		for (final CodingNode node : codingNodes) 
			node.y = 0f;
		
		if (Float.isInfinite(p)) {
			//Winner-take-all
			chooseWinner();
			if (winner!=null)
				winner.y = 1f;
			return;
		}
		
		final float Mf = M;
		final float Tp = Mf - 1e-7f; //Signal close enough to point box at a
		
		//Search for point boxes
		LambdaPrime.clear();
		for (final CodingNode node : Lambda) {
			if ( node.T >= Tp ) 
				LambdaPrime.add(node);
		}
		
		//Calculate activation
		if ( !LambdaPrime.isEmpty() ) {
			//Point box case.
			final float y_pb = 1.0f/LambdaPrime.size();
			for (final CodingNode node : LambdaPrime) {
				node.y = y_pb;
			}
		}
		else { //No point box.
						
			//Calc activity
			float sumy = 0f;
			for (final CodingNode node : Lambda) {
				node.y = pow.calc( Mf - node.T ); // [M-T]^-p
				sumy += node.y;
			}
			final float invSumy = 1f/sumy;
			for (final CodingNode node : Lambda)
				node.y *= invSumy;
			
		}
	}
	
	/**
	 * F2->Fab signal
	 * sigma_k = sum_{j in Lambda with kappa(j)=k} y_j 
	 */
	protected void activateMappingField() {
		
		for (int k=0; k<L; k++) sigma[k] = 0f;
		
		for (final CodingNode node : Lambda)
			sigma[node.kappa] += node.y;
	}
	
	/**
	 * Returns a class prediction K' = argmax_k sigma_k,
	 * breaking ties randomly
	 * @return
	 */
	protected int makePrediction() {
		
		float maxSigma = 0f;
		mappingTies.clear();
		
		for( int k=0; k<L; k++ ) {
			
			if ( sigma[k] > maxSigma ) {
				mappingTies.clear();
				mappingTies.add(k);
				maxSigma = sigma[k];
			}
			else if ( maxSigma-sigma[k] < tieThreshold ) {
				mappingTies.add(k);
			}
			
		}
		
		if (mappingTies.isEmpty())
			return -1;
		
		final int tK = (mappingTies.size()==1) ? 0 : tieBreaker.nextInt(mappingTies.size()); 
		
		return mappingTies.get(tK);
	}
	
	/**
	 * Coding (F2) node, capturing all long-term memory
	 */
	protected static class CodingNode implements Serializable {
		
		private static final long serialVersionUID = -4197320015448919922L;
		
		//LTM
		
		/** Bottom-up coding weights (w_ij) */
		protected FloatVector w;
		
		/** Class membership */
		protected int kappa;
		
		/** Fraction of node undercommitment */
		protected float phi;
		
		
		//TRANSIENTS
		
		/** Match, aka, phasic component */
		protected transient float S;
		
		/** Input */
		protected transient float T;
		
		/** Activation */
		protected transient float y;
		
		/** True if node is in Lambda. */
		protected transient boolean isActive;
		
	}
	
	/**
	 * Provides an iterator over active nodes (node.isActive==true), i.e., Lambda
	 */
	protected class ActiveCodingNodes implements Iterable<CodingNode> {

		public Iterator<CodingNode> iterator() {
			return new Iterator<CodingNode>() {

				private transient final Iterator<CodingNode> src = codingNodes.iterator();
				private transient CodingNode next = null;
				
				public boolean hasNext() {
					if (next!=null)
						return true;
					while( src.hasNext() ) {
						final CodingNode candidate = src.next();
						if (candidate.isActive) {
							next = candidate;
							return true;
						}
					}
					return false;
				}
			
				public CodingNode next() {
					
					if (!hasNext()) 
						throw new NoSuchElementException();
					
					final CodingNode ret = next;
					next = null;
					return ret;
					
				}
			
				public void remove() { src.remove(); }
				
			};
		}
		
	}
	
	/**
	 * Checks to see if all the model parameters have valid values.
	 */
	protected void validateParameters() {
		
		validateParamRange("M",M,1,Integer.MAX_VALUE); 
		validateParamRange("L",L,1,Integer.MAX_VALUE); 
		
		//All others have valid default values. 
		//They can only be changed by their accessor methods,
		//which do the validating, so we don't have to here.
	}
	
	protected void validateParamRange( final String paramName, 
			final float value, final float lb, final float ub )
	{
		if (value<lb || value>ub)
			throw new IllegalParameterException(paramName,value,lb,ub);
	}
	
	/**
	 * Checks to make sure that the passed input conforms to ARTMAP's expectations:
	 * All features must be between 0 and 1.
	 * 
	 * @param input
	 */
	public void validateInput( Input input ) {
		final float[] a = input.features();
		
		if (a.length!=M)
			InputFeaturesLengthMismatchException.assertMatchingNumberOfInputs(M,a.length);
		
		int i = 0;
		for (final float ai : a) {
			if (!Float.isNaN(ai) && (ai<0f || ai>1f))
				throw new InputFeatureOutOfARTBoundsException(i,ai);
			i++;
		}
	}
	/**
	 * Checks to make sure that the passed input conforms to ARTMAP's expectations:
	 * All features must be between 0 and 1, and the class label must be
	 * between 0 and L-1.
	 * 
	 * @param input
	 */
	public void validateLabeledInput(final LabeledInput input) {
		validateInput(input);
		if (!input.isLabeled()) {
			throw new LabelRequiredException("To perform self-supervised learning " +
					"on unlabeled inputs, use selfSupervise(Input).");
		}
		final int K = input.label();
		if (K<0 || K>=getL())
			throw new ClassLabelOutOfBoundsException(0,L,K);
	}
	
	//THE REST ARE ACCESSOR METHODS
	// some with validation code.

	public int getM() {
		return M;
	}

	public void setM(int m) {
		M = m;
	}

	public int getL() {
		return L;
	}

	public void setL(int l) {
		L = l;
	}
	
	public int getE1() {
		return E1;
	}
	public void setE1(int e1) {
		validateParamRange("E1",e1,1,Float.POSITIVE_INFINITY); 
		E1 = e1;
	}
	
	public int getE2() {
		return E2;
	}
	public void setE2(int e2) {
		validateParamRange("E2",e2,1,Float.POSITIVE_INFINITY); 
		E2 = e2;
	}

	public float getAlpha() {
		return alpha;
	}

	public void setAlpha(float alpha) {
		validateParamRange("alpha",alpha,1e-7f,1f-1e-7f); 
		this.alpha = alpha;
	}

	public float getRhobar() {
		return rhobar;
	}

	public void setRhobar(float rhobar) {
		validateParamRange("rhobar",rhobar,0,1); 
		this.rhobar = rhobar;
	}

	public float getBeta1() {
		return beta1;
	}

	public void setBeta1(float beta1) {
		validateParamRange("beta1",beta1,1e-7f,1); 
		this.beta1 = beta1;
	}

	public float getBeta2() {
		return beta2;
	}

	public void setBeta2(float beta2) {
		validateParamRange("beta2",beta2,0f,1f);
		this.beta2 = beta2;
	}

	public float getEpsilon() {
		return epsilon;
	}

	public void setEpsilon(float epsilon) {
		validateParamRange("epsilon",epsilon,-1f+1e-7f,1f-1e-7f);
		this.epsilon = epsilon;
	}
	
	public float getP() {
		return p;
	}
	public void setP(float p) {
		validateParamRange("p",p,1e-7f,Float.POSITIVE_INFINITY);
		this.p = p;
	}
	
	public long getTieBreakerSeed() {
		return tieBreakerSeed;
	}

	public void setTieBreakerSeed(long tieBreakerSeed) {
		this.tieBreakerSeed = tieBreakerSeed;
	}
	
	public float getTieThreshold() {
		return tieThreshold;
	}
	
	public int getNumberOfClasses() { return getL(); }
	public int getNumberOfFeatures() { return getM(); }
	public void setNumberOfClasses(int L) { setL(L); }
	public void setNumberOfFeatures(int M) { setM(M); }

	
	public int C() {
		return codingNodes.size();
	}

	public float[][] w() {
		final float[][] w = new float[codingNodes.size()][];
		int j = 0;
		for (final CodingNode node : codingNodes)
			w[j++] = node.w.toArray();
		return w;
	}
	
	public float[] Phi() {
		final float[] Phi = new float[codingNodes.size()];
		int j = 0;
		for (final CodingNode node : codingNodes)
			Phi[j++] = node.phi;
		return Phi;
	}

	public float[] T() {
		final float[] T = new float[codingNodes.size()];
		int j = 0;
		for (final CodingNode node : codingNodes)
			T[j++] = node.T;
		return T;
	}
	
	public float[] y() {
		final float[] y = new float[codingNodes.size()];
		int j = 0;
		for (final CodingNode node : codingNodes)
			y[j++] = node.y;
		return y;
	}

	public int[] kappa() {
		final int[] kappa = new int[codingNodes.size()];
		int j = 0;
		for (final CodingNode node : codingNodes)
			kappa[j++] = node.kappa;
		return kappa;
	}
	
	@Override
	public String toString() {
		return "SS-AM:"+getParams().toString();
	}
	
	public Map<String,Object> getParams() {
		final Map<String,Object> params = new LinkedHashMap<String,Object>();
		
		params.put("alpha", alpha);
		params.put("beta1",beta1);
		params.put("beta2",beta2);
		params.put("epsilon",epsilon);
		params.put("E1",E1);
		params.put("E2",E2);
		params.put("rhobar",rhobar);
		params.put("p", p);
		
		return params;
		
	}	
	

}
