/**
 *	FloatVector.java
 *
 *  Greg Amis
 *  Technology Lab
 *  Dept. of Cognitive & Neural Systems
 *  Boston University
 *  Oct 9, 2008
 *
 *  Licensing Policy
 *  This implementation is made available to the general public under the terms 
 *  of Copyleft, as defined by the <a href="http://www.fsf.org/">Free Software Foundation</a> 
 *  (see discussion <a href="http://www.gnu.org/copyleft/copyleft.html">here</a> for example). 
 *  As such, it is available free of charge, and may be copied, redistributed, or modified, 
 *  as long as any resulting work is distributed under the same terms of Copyleft. 
 */
package edu.bu.cns.util;

import java.util.Arrays;


/**
 * Fixed-size vector with single floating-point precision.
 * The main point of this is to provide a cache of the vector sum
 * that cannot become accidentally out of sync as changes to 
 * the vector elements are made
 */
public final class FloatVector implements Cloneable, java.io.Serializable {
	
	private static final long serialVersionUID = 477410556451058508L;
	
	private final float[] f;
	private transient float sumf = 0f;
	
	/**
	 * Create a FloatVector instance of the passed size.
	 * @param size
	 */
	public FloatVector(final int size) {
		f = new float[size];
	}
	
	/**
	 * Return the float value at position i
	 * @param i
	 * @return
	 */
	public final float get(final int i) { 
		return f[i];
	}
	
	/**
	 * Set the passed float value to position i,
	 * updating the vector sum.
	 * @param i
	 * @param fi
	 */
	public final void set(final int i, final float fi) { 
		sumf -= f[i];
		f[i] = fi; 
		sumf += fi;
	}
	
	/**
	 * Adds d to the value at position i, returning the new value
	 * @param i
	 * @param d
	 * @return
	 */
	public final float addTo(final int i, final float d) { 
		f[i] += d;
		sumf += d;
		return f[i];
	}
	
	public final float sum() { 
		return sumf; 
	}
	
	public final int size() { 
		return f.length;
	}
	
	public float[] toArray() {
		final float[] copy = new float[f.length];
		System.arraycopy(f, 0, copy, 0, f.length);
		return copy;
	}
	
	/**
	 * Copy array g into FloatVector, starting at position pos.
	 * @param g
	 * @param pos
	 */
	public final void copyInto(final float[] g, final int pos) {
		
		int i = pos;
		for (final float gi : g)
			set(i++,gi);
			
	}
	
	
	@Override
	public FloatVector clone() {
		final FloatVector copy = new FloatVector(this.size());
		copy.copyInto(this.f,0);
		return copy;
	}

    /**
     * Reconstitute the <tt>FloatVector</tt> instance from a stream (that is,
     * deserialize it). Update sumf, which wouldn't have been serialized.
     */
    private final void readObject(final java.io.ObjectInputStream s)
        throws java.io.IOException, ClassNotFoundException {
		s.defaultReadObject();
		sumf = 0f;
		for ( final float fi : f) sumf += fi;
    }
    
    @Override
    public String toString() {
    	return Arrays.toString(f);
    }

}
