/**
 *	ForkProcessor.java
 *
 *  Greg Amis
 *  Technology Lab
 *  Dept. of Cognitive & Neural Systems
 *  Boston University
 *  Apr 4, 2008
 *
 *  Licensing Policy
 *  This implementation is made available to the general public under the terms 
 *  of Copyleft, as defined by the <a href="http://www.fsf.org/">Free Software Foundation</a> 
 *  (see discussion <a href="http://www.gnu.org/copyleft/copyleft.html">here</a> for example). 
 *  As such, it is available free of charge, and may be copied, redistributed, or modified, 
 *  as long as any resulting work is distributed under the same terms of Copyleft.
 */
package edu.bu.cns.util;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.FutureTask;

/**
 * Handles fork-and-join paralellism.
 * 
 * Usage:
 * 		forkProcessor = new ForkProcessor<int[]>(numTines);
 * 		while (more tines)
 * 			forkProcessor.submit( tineTask );
 * 		results = forkProcessor.finish();
 */
public class ForkProcessor<T> {
	
	public static final <V> List<V> processAll( final Collection<Callable<V>> cs ) {
		final ForkProcessor<V> fp = new ForkProcessor<V>(cs.size());
		for ( final Callable<V> c : cs )
			fp.submit(c);
		return fp.finish();
	}
	
	private final List<FutureTask<T>> futures;
	
	public ForkProcessor(final int numTines) {
		futures = new ArrayList<FutureTask<T>>(numTines);
	}
	
	public void submit(Callable<T> c) {
		futures.add( (FutureTask<T>) threadPool.submit(c) );
	}
	
	public boolean isCanceled() { return false; }
	
	public List<T> finish() {
		
		final List<T> result = new ArrayList<T>(futures.size());
		
		for (final FutureTask<T> future : futures) {	
			if (isCanceled())
				break;			//Run any tasks not handled by the threadPool
			future.run();		//with the current thread
		}
		
		try {
			for (final FutureTask<T> future : futures) {
				if (isCanceled() && !future.isDone()) {
					future.cancel(true);
					result.add(null);
					continue;
				}
				result.add( future.get() );
			}
		} catch (Exception e) {
			throw new RuntimeException(e);
		}
		finally {
			futures.clear();
		}
		
		return result;
	}
	
	private static final ExecutorService threadPool 
		= Executors.newFixedThreadPool( Math.max( 1,
				Runtime.getRuntime().availableProcessors() - 2 ) );

}
